/*
 * ProFTPD: mod_gzipfs -- an FS module for automatically compressing and
 *             uncompressing files
 *
 * Copyright (c) 2001-2003 TJ Saunders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exemption, TJ Saunders gives permission to link this program
 * with OpenSSL, and distribute the resulting executable, without including
 * the source code for OpenSSL in the source distribution.
 *
 * This is mod_gzipfs, contrib software for proftpd 1.2.2 and above.
 * For more information contact TJ Saunders <tj@castaglia.org>.
 *
 * NOTE:  until a $Includes$ is supported, the proper --with-includes
 *  configure option will be needed for this.
 *
 * -- DO NOT MODIFY THE LINES BELOW --
 * $Libraries: -lz$
 * $Id: mod_gzipfs.c,v 1.15 2004/02/05 18:19:26 tj Exp tj $
 */

#include "conf.h"

#define MOD_GZIPFS_VERSION "mod_gzipfs/0.9rc4"

/* Make sure the version of proftpd is as necessary. */
#if PROFTPD_VERSION_NUMBER < 0x0001020901
# error "ProFTPD 1.2.9rc1 or later required"
#endif

#include "zlib.h"

/* Fake file descriptor value for this module, for emulating library
 * filesystem functions
 */
#define GZIPFS_FAKE_FD	7

/* Scratchwork buffer */
#define GZIPFS_BUFLEN	1024

static char scratchpad[GZIPFS_BUFLEN];

/* Magic header for gzip files, 1F 8B */
static char gzipfs_magic[2] = {'\037', '\213'};

/* Flag for signalling when to examine a stream for the identifying magic */
static int chk_stream_compression = FALSE;

/* A struct for carrying around zlib data.  Note that at present, this
 * single structure is used for the mod_gzipfs pr_fs_t, meaning that it
 * _will_ run into reentrancy issues
 */
struct gzipfs_data {

  /* Pool for this object's use */
  pool *gzpool;

  /* Configured compression level */
  int gzcompression;

  /* zlib file pointer */
  gzFile *gzfile;

  /* Path to opened gzipfs file */
  char *gzfile_path;

  /* Flags for opened gzipfs file */
  int gzfile_flags;

  /* Backup file descriptor, for uploaded files that are already compressed. */
  int compressedfd;
};

static array_header *gzipfs_list = NULL;

/* Support routines
 */

static char *add_gzipfs_extension(const char *path) {
  char *charp = NULL;

  /* Clear the scratchpad.  Automatically add a ".gz" extension to
   * the path being opened.
   */
  memset(scratchpad, '\0', GZIPFS_BUFLEN);
  sstrcat(scratchpad, path, GZIPFS_BUFLEN);

  if (strcmp(path, ".") == 0 ||
      strcmp(path, "..") == 0)
    return scratchpad;

  /* Watch out for paths that already have the ".gz" extension.
   * As there isn't a nice strrstr() library function, this is a little
   * messy.  Find the end of the given path, back up three characters, and
   * match strings.
   */
  charp = &(scratchpad[strlen(scratchpad) - 3]);
  if (strcmp(charp, ".gz") == 0)
    return scratchpad;

  sstrcat(scratchpad, ".gz", GZIPFS_BUFLEN);

  return scratchpad;
}

static char *strip_gzipfs_extension(const char *path) {
  char *charp = NULL;

  memset(scratchpad, '\0', GZIPFS_BUFLEN);
  sstrcat(scratchpad, path, GZIPFS_BUFLEN);

  charp = &(scratchpad[strlen(scratchpad) - 3]);
  if (strcmp(charp, ".gz") == 0)
    *charp = '\0';

  return scratchpad;
}

/* This function is necessary to keep the server from dying upon a SIGHUP.
 * The server, once restarted, will re-call all of the configuration directive
 * handlers, and in this module's case, this will cause the handler to try
 * to re-register a path already present in the FS layer's map -- something
 * that is currently not supported.  As a workaround, this module keeps
 * track of all paths it registers, and upon receiving a SIGHUP, it removes
 * the paths from the FS layer's map before the configuration directive
 * handlers are called.
 */
static void rehash_gzipfs(void *data) {
  register unsigned int i = 0;
  pool *list_pool = NULL;
  char **gzipfs_objs = (char **) gzipfs_list->elts;

  for (i = 0; i < gzipfs_list->nelts; i++)
    pr_unregister_fs(gzipfs_objs[i]);

  destroy_pool(gzipfs_list->pool);

  list_pool = make_sub_pool(permanent_pool);
  gzipfs_list = make_array(list_pool, 0, sizeof(char *));

  return;
}

/* Prepare the module-specific list of registered gzipfs paths, and
 * register the SIGHUP handler
 */
static int init_gzipfs(void) {
  pool *list_pool = make_sub_pool(permanent_pool);
  gzipfs_list = make_array(list_pool, 0, sizeof(char *));
  pr_rehash_register_handler(NULL, rehash_gzipfs);
  return 0;
}

static int is_compressed_stream(const char *buf, size_t buflen) {

  /* ASSUME that the input buffer is of the very first bytes in the
   * stream, in which case we need only examine the first two bytes
   */
  if (buf[0] == gzipfs_magic[0] &&
      buf[1] == gzipfs_magic[1])
    return TRUE;
  
  return FALSE;
}

/* FSIO callbacks */

static int gzipfs_close(pr_fh_t *fh, int fd) {
  int res;
  struct gzipfs_data *data = fh->fh_data;

  if (data->gzfile != NULL) {
    gzFile *gzfile = NULL;

    /* NOTE: how to handle FS modules such as this, which may return error
     *  values specific to the module, and not relevant to strerror()?
     */

    /* Retrieve the gzFile pointer */
    gzfile = data->gzfile;

    res = gzclose(gzfile);
    if (res != 0) {
      if (res == Z_ERRNO) {
        return res;

      } else {
        log_pri(LOG_ERR, "error: mod_gzipfs close(): %s", gzerror(gzfile,
          &res));

        /* Fake an errno value, knowing that the real cause of the problem
         * is logged.
         */
        errno = EPERM;
        return -1;
      }
    }

    data->gzfile = NULL;

  } else {

    /* Clean up. NOTE: in 1.3, this pool can be cleared instead of
     * reallocated
     */

    data->gzfile_path = NULL;
    destroy_pool(data->gzpool);
    data->gzpool = make_sub_pool(fh->fh_pool);

    /* Just a simple close will do */
    return close(data->compressedfd);
  }

  return 0;
}

/* This function is needed for files retrieved using REST.  Warning, as
 * per the cautionary notes in zlib.h, use of this function can be
 * extremely slow
 */
static off_t gzipfs_lseek(pr_fh_t *fh, int fd, off_t offset, int whence) {
  gzFile *gzfile = NULL;

  /* Retrieve the gzFile pointer */
  gzfile = ((struct gzipfs_data *) fh->fh_fs->fs_data)->gzfile;

  return gzseek(gzfile, offset, whence);
}

static int gzipfs_lstat(pr_fs_t *fs, const char *path, struct stat *sbuf) {
  int res = lstat(add_gzipfs_extension(path), sbuf);

  if (res == -1)
    return lstat(path, sbuf);

  return res;
}

static int gzipfs_open(pr_fh_t *fh, const char *path, int flags) {
  char mode[4] = {'\0'};
  gzFile gzfile = NULL;
  struct gzipfs_data *data;

  data = pcalloc(fh->fh_pool, sizeof(struct gzipfs_data));

  /* This file uses the compression value stored in fs_data. */
  data->gzcompression = *((int *) fh->fh_fs->fs_data);
  data->gzfile = NULL;
  data->gzpool = make_sub_pool(fh->fh_pool);
  data->gzfile_flags = flags;

  /* Check for O_RDONLY, O_WRONLY, or O_RDWR in flags, and determine the
   * mode to pass to gzopen().
   */
  if (flags & O_WRONLY) {
    if (data->gzcompression)
      snprintf(mode, 4, "wb%u", data->gzcompression);
    else
      sstrcat(mode, "wb", GZIPFS_BUFLEN);

    chk_stream_compression = TRUE;

  } else if (flags & O_RDWR) {
    if (data->gzcompression)
      snprintf(mode, 4, "w+%d", data->gzcompression);
    else
      sstrcat(mode, "w+", GZIPFS_BUFLEN);

    chk_stream_compression = TRUE;

  } else {

    /* Default to opening the file in read-only mode */
    if (data->gzcompression)
      snprintf(mode, 4, "rb%u", data->gzcompression);
    else
      sstrcat(mode, "rb", GZIPFS_BUFLEN);
  }

  /* Store the path and flags for which of the two opens succeeds. */
  
  gzfile = gzopen(add_gzipfs_extension(path), mode);
  if (gzfile == NULL) {
    gzfile = gzopen(path, mode);
    if (gzfile == NULL)
      return -1;
 
    data->gzfile_path = pstrdup(data->gzpool, path);

  } else {
    data->gzfile_path = pstrdup(data->gzpool, add_gzipfs_extension(path));
  }
 
  data->gzfile = gzfile;
  fh->fh_data = data;

  /* Return any non-zero integer, to emulate open(2).  We won't know until
   * the first fs_write() whether to continue using the zlib functions
   * for compression.
   */
  return GZIPFS_FAKE_FD;
}

static int gzipfs_read(pr_fh_t *fh, int fd, char *buf, size_t buflen) {
  gzFile *gzfile = NULL;

  /* Retrieve the gzFile pointer. */
  gzfile = ((struct gzipfs_data *) fh->fh_data)->gzfile;

  /* gzread() handles non-compressed files transparently. */
  return gzread(gzfile, buf, buflen);
}

static int gzipfs_rename(pr_fs_t *fs, const char *rnfm, const char *rnto) {
  int res = 0;

  /* this is needed to catch RNFR, RNTO commands of a file in a GzipDirectory,
   * for which the client can't see the ".gz" extension.  It looks like
   * this function will be passed the full paths.
   *
   * NOTE: how to handle renaming of files from a non-GzipDirectory into a
   *  GzipDirectory, and vice versa?  This may be tricky, depending on
   *  which FS rename() pointer is called.  As things are now, this
   *  is most likely going to be called on to handle three situations:
   *  1) the rnfm filename is handled by this FS module, as is the rnto
   *  filename, 2) the rnfm filename is handled by this FS module, but
   *  the rnto filename is not, or 3) neither rnfm nor rnto filenames are
   *  handled by this FS module, it just happens that the current
   *  working of the directory is in this FS module's demesne when the
   *  client calls to rename files outside the FS module's demesne.
   */

  if (!strncmp(rnfm, fs->fs_path, strlen(fs->fs_path))) {
    if (!strncmp(rnto, fs->fs_path, strlen(fs->fs_path))) {

      /* a temporary pool is needed here, as add_gzipfs_extension() uses
       * a static buffer, and calling it twice would overwrite the
       * contents of that buffer; the pool is used to store temporary
       * adjustments.
       */
      pool *tmp_pool = make_sub_pool(NULL);
      char *new_rnfm = pstrdup(tmp_pool, add_gzipfs_extension(rnfm));
      char *new_rnto = pstrdup(tmp_pool, add_gzipfs_extension(rnto));

      res = rename(new_rnfm, new_rnto);

      destroy_pool(tmp_pool);
 
    } else {
      res = rename(add_gzipfs_extension(rnfm), rnto);
    }

  } else {
    if (!strncmp(rnto, fs->fs_path, strlen(fs->fs_path))) {
      res = rename(rnfm, add_gzipfs_extension(rnto));
    } else {
      res = rename(rnfm, rnto);
    }
  }

  return res; 
}

static int gzipfs_stat(pr_fs_t *fs, const char *path, struct stat *sbuf) {
  int res = stat(add_gzipfs_extension(path), sbuf);

  if (res == -1)
    return stat(path, sbuf);

  return res;
}

static int gzipfs_unlink(pr_fs_t *fs, const char *path) {
  int res;

  res = unlink(add_gzipfs_extension(path));

  if (res == -1 && errno == ENOENT)
    return unlink(path);

  return res;
}

static int gzipfs_write(pr_fh_t *fh, int fd, const char *buf, size_t buflen) {
  gzFile *gzfile = NULL;
  struct gzipfs_data *data = fh->fh_data;

  /* Retrieve the gzFile pointer */
  gzfile = data->gzfile;

  /* How to handle the uploading of files which may already be compressed?
   *
   * To determine this, the chk_stream_compression flag is TRUE whenever
   * files are opened for reading in a mod_gzipfs directory.  So, on the
   * initial write, check for the magic bytes that signal whether the
   * incoming stream is already compressed.  If not, do nothing.  If so, then
   * things become a little more interesting: instead of using the
   * zlib functions, gzclose() and unlink() the already open file, then
   * open and write to another, behind the scenes.  What a pain.
   */

  if (chk_stream_compression) {
    if (is_compressed_stream(buf, buflen)) {

      /* Clear the stream check flag. */
      chk_stream_compression = FALSE;

      gzclose(gzfile);
      data->gzfile = NULL;

      unlink(data->gzfile_path);

      data->compressedfd = open(data->gzfile_path, data->gzfile_flags,
        PR_OPEN_MODE);
    }
  }

  if (data->gzfile != NULL)
    return gzwrite(gzfile, (void *) buf, buflen);

  else 
    return write(data->compressedfd, buf, buflen);

  return 0;
}

static struct dirent *gzipfs_readdir(pr_fs_t *fs, void *dir) {
  struct dirent *dir_entry = NULL;

  /* This function is needed to hide the ".gz" extensions of all the files
   * in the GzipDirectory, to keep things transparent to the FTP client
   * (yet still allow non-ProFTPD utilities to know what those files are).
   */

  if ((dir_entry = readdir((DIR *) dir)) != NULL) {
    /* Replace the filename in the dirent struct with the gzipfs version. */
    sstrncpy(dir_entry->d_name, strip_gzipfs_extension(dir_entry->d_name),
      sizeof(dir_entry->d_name));

  /* NOTE: what happens if there is a "file" and a "file.gz"?  It will be
   *  listed twice as "file" -- how to handle this cleanly?
   */
  }

  return dir_entry;
}

static int add_gzipfs_path(pool *p, const char *path, int compression) {
  pr_fs_t *fs = NULL;

  fs = pr_register_fs(p, "mod_gzipfs", path);
  if (fs == NULL)
    return FALSE;

  /* Add the custom zlib FSIO functions here.  All other non-specified
   * functions will default to the system version, which is fine.
   */
  fs->close = gzipfs_close;
  fs->lseek = gzipfs_lseek;
  fs->lstat = gzipfs_lstat;
  fs->open = gzipfs_open;
  fs->read = gzipfs_read;
  fs->rename = gzipfs_rename;
  fs->stat = gzipfs_stat;
  fs->unlink = gzipfs_unlink;
  fs->write = gzipfs_write;
  fs->readdir = gzipfs_readdir;

  /* Instantiate an object for module-specific data, and store it in
   * pr_fs_t's data field.
   */
  fs->fs_data = pcalloc(fs->fs_pool, sizeof(int)); 
  *((int *) fs->fs_data) = compression; 

  /* Add the registered path to this module's internal list. */
  *((char **) push_array(gzipfs_list)) = pstrdup(gzipfs_list->pool, path);

  return TRUE;
}

/* configuration directive handlers */

MODRET set_gzipfs(cmd_rec *cmd) {
  int compression = 0;
  char *path = NULL;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_ANON|CONF_VIRTUAL|CONF_GLOBAL);

  /* the compression factor is an optional argument */
  if (cmd->argc-1 == 2) {
    compression = atoi(cmd->argv[2]);

    /* check that compression is an allowed number (0-9) */
    if (compression < 0 || compression > 9)
      return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
        cmd->argv[0], ": compression must be 0-9", NULL));
  }

  /* make sure this is an acceptable path */
  if ((cmd->argv[1])[0] == '.' && (cmd->argv[1])[1] == '.')
    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "'..' prefix not allowed in path: ",
    cmd->argv[1], NULL));

  if (strchr(cmd->argv[1], '*'))
    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "wildcards not allowed in path: ",
    cmd->argv[1], NULL));

  /* make sure the given path ends in a slash -- very important! */
  path = cmd->argv[1];
  if (path[strlen(path) - 1] != '/')
    path = pstrcat(cmd->tmp_pool, path, "/", NULL);

  if (add_gzipfs_path(cmd->server->pool, path, compression)) {
    log_debug(DEBUG5, "GzipFS initialized for %s", path);

  } else {
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      cmd->argv[0], ": error initializing GzipFS for '", path,
      "'", NULL));
  }
 
  return HANDLED(cmd);
}

static conftable gzipfs_conftab[] = {
  { "GzipFS",	set_gzipfs,	NULL },
  { NULL }
};

module gzipfs_module = {

  /* always NULL */
  NULL, NULL,

  /* Module API version 2.0 */
  0x20,

  /* the module name */
  "gzipfs",

  /* module configuration handler table */
  gzipfs_conftab,

  /* module command handler table */
  NULL,

  /* module authentication handler table */
  NULL,

  /* module initialization function */
  init_gzipfs,

  /* module "child mode" post-fork initialization function */
  NULL
};
